# -*- coding: utf-8 -*-

class VraSdkException(Exception):
    """Main class exception"""
    pass

class VraSdkRequestException(VraSdkException):
    """for requests.exceptions.RequestException"""
    pass

class VraSdkAuthenticateException(VraSdkException):
    """for vra_authenticate"""
    pass

class VraSdkConfigException(VraSdkException):
    """for explicit configuration file issue"""
    pass

class VraSdkMainConfigException(VraSdkException):
    """for vra_config"""
    pass

class VraSdkMainException(VraSdkException):
    """for vra_sdk"""
    pass

class VraSdkFactoryException(VraSdkException):
    """for vra_factory"""
    pass

class VraSdkEntitlementException(VraSdkException):
    """for explicit entitlement"""
    pass

class VraSdkPayloadException(VraSdkException):
    """for vra_payload_*"""
    pass

class VraSdkUtilsException(VraSdkException):
    """for vra_utils"""
    pass

class VraSdkDecoratorException(VraSdkException):
    """for vra_decorator"""
    pass

class VraSdkMainRequestException(VraSdkException):
    """for vra_request"""
    pass


# -*- coding: utf-8 -*-
import requests
import urllib3
import re
import json
import time
import importlib
from vra_sdk.vra_formatter import format_result
from vra_sdk.vra_utils import get_module_class
from vra_sdk.vra_authenticate import VraConfig
from vra_sdk.vra_factory import VraFactory
from vra_sdk.vra_exceptions import VraSdkConfigException, VraSdkRequestException, VraSdkMainRequestException


class VraRequest():
    """Handle getting data from vRa infrastructure as well as catalog item/resource action request

    Attributes:
        config (VraConfig): VraConfig object
        Payload (vra_payload_x.CatalogItem or vra_payload_x.ResourceAction): Payload object
        status_url (string): url to get the status of the current request
        response (requests.Response): request response
    """

    def __init__(self, payload, **kwargs):
        """Init the VraRequest object
        
        Args:
            payload (CatalogItem or ResourceAction): payload object
        """

        urllib3.disable_warnings()
        self.config = VraConfig()
        self.payload = payload
        self.status_url = 'not set'
        self.response = None

    def format_filters(self, object_type, key, value, resource_type=None):
        """Handle generation of OData url filter
        
        Args:
            object_type (string): type of object ot create filter for
            key (string): field to filter from
            value (string): value of the field to filter
            resource_type (string, optional): Defaults to None. Use only for get_raw_definitions(), use to create a filter without defintions
        
        Returns:
            string: Odata filter
        """

        filters_array = []
        computed = []

        if resource_type is None:

            if object_type not in self.config.config_file["business_models"]:
                raise VraSdkConfigException(
                    "Cannot retrieve resource type associated with your object type, have you declared it in config.json map_resource_type array?")

            try:
                my_module, _ = get_module_class(
                    self.config.config_file["business_models"][object_type]["path"])
            except Exception as e:
                raise VraSdkMainRequestException(f"Error importing module.class {self.config.config_file['business_models'][object_type]} Error:{e}")

            for ref in getattr(my_module, 'RESOURCE_TYPE'):
                computed.append(f"(resourceType/name+eq+'{ref}')")
        else:
            computed.append(f"(resourceType/name+eq+'{resource_type}')")

        if computed:
            filters_array.append("(" + "+or+".join(computed) + ")")

        if key and value:
            filters_array.append("(" + key + "+eq+'" + value + "')")

        result = "$filter=" + \
            "(" + "+and+".join(filters_array) + ")" if filters_array else None
        return result

    def get_status(self):
        """Use the status_url attribute to get the status of the request against the vRa infrastructure
        
        Returns:
            dict: vRa status state
        """

        try:
            req = self.config.session.get(
                self.status_url, verify=self.config.verify, timeout=self.config.timeout)
            res = json.loads(req.text)['state']
            return res
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'Error requesting status url {self.status_url}: {e}')
        except Exception as e:
            raise VraSdkMainRequestException(
                f'Unmanaged error requesting status url {self.status_url}: {e}')

    def execute_async(self):
        """Execute the self.payload object execute_request() method. Set the status_url if there's one in the vRa answer.
        
        Returns:
            self: VraRequest
        """

        self.response = self.payload.execute_request()
        if hasattr(self.response, 'headers'):
            self.status_url = self.response.headers.get(
                'Location', 'NoLocationFound')

        return self

    def execute_sync(self):
        """wrapper of execute_async and retry it until the result if successful or failed
        
        Returns:
            self: VraRequest
        """

        self.execute_async()
        while self.get_status() != 'SUCCESSFUL' and self.get_status() != 'PROVIDER_FAILED':
            time.sleep(3)
        if self.get_status() != 'SUCCESSFUL':
            raise VraSdkMainRequestException('Request failed')
        else:
            return self

    def get_object_raw(self, object_type, key, value, limit, page, full=False, resource_type=None):
        """Get raw catalog resource information from vRa infrastructure
        
        Args:
            object_type (string): type of vRa resource to get data on
            key (string): field to search for
            value (string): value of the field
            limit (int): maximum result per page
            page (int): page to get from result.
            full (bool): If True return the full result
            resource_type (string, optional): Defaults to None. Only used for get_raw_definitions()
        
        Returns:
            dict: raw vRa data
        """

        
        #url_array = [f"limit={str(limit)}",f'page={str(page)}']
        url_array = ["limit=" + str(limit), "page=" + str(page)]
        filters = self.format_filters(object_type, key, value, resource_type)
        if filters:
            url_array.append(filters)
        amp = "&"
        url = f"https://{self.config.vcac_server}/catalog-service/api/consumer/resources/?{amp.join(url_array)}"
        try:
            req = self.config.session.get(
                url, verify=self.config.verify, timeout=self.config.timeout)
            req.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkMainRequestException(
                f'Unmanaged error requesting vRa: {e}')

        res = req.json()

        if (not object_type and resource_type) or full:
            ids = []
            if key != 'id':
                for elt in res["content"]:
                    ids.append(elt["id"])
            else:
                ids.append(value)

            result = []
            for id in ids:
                url = f"https://{self.config.vcac_server}/catalog-service/api/consumer/resources/{id}"
                try:
                    req = self.config.session.get(
                        url, verify=self.config.verify, timeout=self.config.timeout)
                    req.raise_for_status()
                except requests.exceptions.RequestException as e:
                    raise VraSdkRequestException(
                        f'vRa request exception : {e}')
                except Exception as e:
                    raise VraSdkMainRequestException(
                        f'Unmanaged error requesting vRa: {e}')

                res = req.json()
                result.append(res)
            return result
        else:
            return res['content'] if 'content' in res else []

    def get_object(self, object_type, key, value, limit, page, full=False, resource_type=None):
        """Get raw_data from get_raw_object() and prettify it to then create a list of object using the factory and these data.
        
        Args:
            object_type (string): type of vRa resource to get data on
            key (string): field to search for
            value (string): value of the field
            limit (int): maximum result per page
            page (int): page to get from result.
            resource_type (string, optional): Defaults to None. Only used for get_raw_definitions()
        
        Returns:
            list: list of object type as defined in the business_models configuration section 
        """

        result = []
        object_result = []
        raw_data = self.get_object_raw(object_type, key, value, limit, page, full, resource_type)
        # Contruct dict of result without raw_data
        if raw_data is not None:
            for elt in raw_data:
                result.append(format_result(elt))
            if key and value:
                result = [obj for obj in result if re.match(
                    value, obj.get(key, ""))]
        if resource_type is not None:
            return result

        # Adding raw_data and contruct object array
        if result:
            for i, elt in enumerate(result):
                elt['raw_data'] = raw_data[i]
                object_result.append(VraFactory.factory(object_type, **elt))
            return object_result

    def get_request_result_raw(self):
        """Request vRa to get the status of a specific request based on the status_url
        
        Returns:
            dict: vRa raw data
        """

        try:
            req = self.config.session.get(f"{self.status_url}/resources", verify=self.config.verify, timeout=self.config.timeout)
            req.raise_for_status()
            data = json.loads(req.text)
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkMainRequestException(
                f'Unmanaged error requesting vRa: {e}')

        if data.get("content"):
            return data['content'][0]
        else:
            try:
                req = self.config.session.get(f"{self.status_url}/forms/details", verify=self.config.verify, timeout=self.config.timeout)
                req.raise_for_status()
            except requests.exceptions.RequestException as e:
                raise VraSdkRequestException(
                    f'vRa request exception : {e}')
            except Exception as e:
                raise VraSdkMainRequestException(
                    f'Unmanaged error requesting vRa: {e}')
            data = json.loads(req.text)
            return data

    def get_request_result(self):
        """wrapper of get_request_result_raw(). Return the beautify result
        
        Returns:
            dict: vRa beautify result
        """

        return format_result(self.get_request_result_raw())



# -*- coding: utf-8 -*-
import os
import re
import sys
import json
from pathlib import Path
import importlib
import time
from vra_sdk.vra_exceptions import VraSdkUtilsException, VraSdkConfigException


def resolve_path(config_path):
    """Return the absolute path of a file
    
    Args:
        config_path (string): absolute or relative path to a file
    
    Returns:
        string: path to the config_path file
    """

    return os.path.abspath(os.path.join(os.getcwd(), config_path))


def prettify_key(initial_dict):
    """transform camelCase dict key to snake_case
    
    Args:
        dict (dict): dict to transform key of
    
    Returns:
        dict: transformed dict
    """

    dict_clean = {}
    for key, value in initial_dict.items():
        dict_clean[to_snake_case(key)] = value
    return dict_clean


def to_snake_case(name):
    """transform CamelCase string to snake_case
    
    Args:
        name (string): string to convert
    
    Returns:
        string: converted string
    """

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



def get_module_class(kls):
    """Provide module and class from a string
    
    Args:
        kls (string): path to the module
    
    Returns:
        tuple: tuple of module, class
    """

    mod, _, cls = kls.rpartition('.')
    mod = importlib.import_module(mod)

    return mod, getattr(mod, cls)


def clean_kwargs_key(**kwargs):
    """replace dot (.) to underscore for kwargs key
    
    Returns:
        dict: transformed kwargs
    """

    cleaned_kwargs = {k.replace('.', '_'): v for (k, v) in kwargs.items()}
    return cleaned_kwargs


def load_payload_file(payload_path):
    """load json file
    
    Args:
        payload_path (string): path to the json file
    
    Returns:
        dict: serialization of the json file
    """

    try:
        with open(resolve_path(payload_path)) as f:
            base = json.load(f)
        return base
    except FileNotFoundError as e:
        raise VraSdkConfigException(
            f'Error opening payload {payload_path}. Does the file exist?: {e}')
    except Exception as e:
        raise VraSdkUtilsException(
            f'Unmanaged error during payload loading for payload {payload_path}: {e}')



# -*- coding: utf-8 -*-
from vra_sdk.models.vra_object import VraBaseObject
from vra_sdk.vra_config import VraConfig
from vra_sdk.vra_utils import load_payload_file
from vra_sdk.vra_exceptions import VraSdkRequestException, VraSdkPayloadException
import requests
import json
from copy import deepcopy 

class BasePayload():
    """Base class for CatalogItem and ResourceAction
    """

    def __init__(self):
        self.config = VraConfig()

    def customize_payload(self, payload, **kwargs):
        """base customization payload
        
        Args:
            payload (dict): base payload, without any customization
        
        Returns:
            dict: customized payload
        """

        if kwargs['payload_type'] == 'CatalogItem':
            payload['catalogItemId'] = kwargs['catalog_item_id']
            payload['requestedFor'] = kwargs['requested_for']
            payload['businessGroupId'] = kwargs['business_group_id']
        else:
            payload['actionId'] = kwargs['resource_action_id']
            payload['resourceId'] = kwargs['resource_id']
            if 'description' in kwargs:
                payload['description'] = kwargs['description']

        key_to_performed = [key for key in kwargs.keys() if key not in self.config.config_file['not_in_data']]
        for key in key_to_performed:
            payload['data'][key] = kwargs[key]

        return payload


class CatalogItem(BasePayload):

    def __init__(self, customization_func=None, **kwargs):
        """Init ResourceAction object for vRa 7.x payload object
            customization_func ([type], optional): Defaults to None. If not None, this function will add a second customization after the initial one.
        """

        super().__init__()
        self.customized = None
        if not kwargs.get('payload_path'):
            self.base = self.get_template(kwargs.get('catalog_item_id'))
        else:
            self.base = load_payload_file(kwargs['payload_path'])

        self.customized = self.customize_payload(deepcopy(self.base), **kwargs)
        if customization_func:
            self.customized = customization_func(self.customized, **kwargs)

    def execute_request(self):
        """Execute the request against the vRa infrastructure
        
        Returns:
            requests.Request: request object
        """

        try:
            req = self.config.session.post(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/entitledCatalogItems/{self.customized['catalogItemId']}/requests",
                json=self.customized,
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
            return req
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')

    def get_template(self, catalog_item_id):
        """Get payload template for catalog item request against vRa infrastructure
        
        Args:
            catalog_item_id (string): id of the catalog item
        
        Returns:
            dict: payload of the request to perform to request the specified catalog item
        """

        try:
            req = self.config.session.get(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/entitledCatalogItems/{catalog_item_id}/requests/template",
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
            return json.loads(req.text) 
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')
        

class ResourceAction(BasePayload):
    def __init__(self, customization_func=None, **kwargs):
        """Init ResourceAction object for vRa 7.x payload object
            customization_func ([type], optional): Defaults to None. If not None, this function will add a second customization after the initial one.
        """

        super().__init__()
        self.customized = None
        if not kwargs.get('payload_path'):
            self.base = self.get_template(kwargs.get('resource_id'), kwargs.get('resource_action_id'))
        else:
            self.base = load_payload_file(kwargs['payload_path'])

        self.customized = self.customize_payload(deepcopy(self.base), **kwargs)
        if customization_func:
            self.customized = customization_func(self.customized, **kwargs)

    def execute_request(self):
        """Execute the request against the vRa infrastructure
        
        Returns:
            requests.Request: request object
        """

        try:
            req = self.config.session.post(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/resources/{self.customized['resourceId']}/actions/{self.customized['actionId']}/requests",
                json=self.customized,
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
            return req
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')

    def get_template(self, resource_id, resource_action_id):
        """Get payload template for resource action request against vRa infrastructure
        
        Args:
            resource_id (string): id of the resource to perform the action on
            resource_action_id (string): id of the action to perform
        
        Returns:
            dict: payload of the request to perform to request the specified action on the specified resource
        """

        try:
            req = self.config.session.get(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/resources/{resource_id}/actions/{resource_action_id}/requests/template",
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
            return json.loads(req.text)
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')



# -*- coding: utf-8 -*-
import json
import sys
from dateutil.parser import parse

from vra_sdk import vra_utils


def parse_string(element):
    """parser of vRa data of string type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key,value of vRa data
    """

    return element["key"], element["value"]["value"]


def parse_string_without_key(element):
    """parser of vRa data without key
    
    Args:
        element (dict): vRa data
    
    Returns:
        string: vRa return value
    """

    return element["value"]


def parse_integer(element):
    """parser of vRa data integer type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key,value of vRa data
    """

    return element["key"], int(element["value"]["value"])


def parse_decimal(element):
    """parser of vRa data decimal type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key, value of vRa data
    """

    return element["key"], float(element["value"]["value"])


def parse_boolean(element):
    """parser for vRa data of boolean type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key, value of vRa data
    """

    return element["key"], bool(element["value"]["value"])


def parse_datetime(element):
    """parser for vRa data datetime type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key, datatime of vRa data
    """

    return element["key"], parse(element["value"]["value"]).strftime("%Y-%m-%dT%H:%M:%S.000+0100")


def parse_multiple(element):
    """parser for vRa data multiple type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key, value of vRa data
    """

    parsed = []
    for elt in element["value"]["items"]:
        try:
            value = parse_key(elt)[1]
        except TypeError:
            value = parse_string_without_key(elt)
        parsed.append(value)
    return element["key"], parsed


def parse_complex(element):
    """parser for vRa data complex type
    
    Args:
        element (dict): vRa data
    
    Returns:
        tuple: key,value of vRa data
    """

    parsed = {}
    if "values" in element:
        entries = element["values"]["entries"]
    else:
        entries = element["value"]["values"]["entries"]

    for elt in entries:
        key, value = parse_key(elt)
        parsed[key] = value

    key = element["key"] if "key" in element else "0"
    return key, parsed


def parse_key(element):
    """proxy  method to class the specific parser
    
    Args:
        element (dict): vRa result element
    
    Returns:
        func: parser function to call
    """

    elt_type = element["value"]["type"] if "value" in element else element["type"]
    return getattr(sys.modules[__name__], "parse_%s" % elt_type.lower())(element)


def format_result(raw_result):
    """Format raw vRa result to a more user friendly result
    
    Args:
        raw_result (dict): raw data of vRa infrastructure when getting data
    
    Returns:
        dict: user friendly vRa result
    """

    data_clean = {}
    if 'id' in raw_result:
        data_clean["id"] = raw_result['id']

    if 'name' in raw_result:
        data_clean["name"] = raw_result['name']

    if 'status' in raw_result:
        data_clean["status"] = raw_result['status']

    if 'lease' in raw_result:
        data_clean["lease"] = raw_result['lease']

    if 'description' in raw_result:
        data_clean["description"] = raw_result['description']

    if 'organization' in raw_result:
        data_clean["business_group"] = {}
        data_clean["business_group"]["id"] = raw_result["organization"]["subtenantRef"]
        data_clean["business_group"]["label"] = raw_result["organization"]["subtenantLabel"]

    value_to_iter_on = "values" if "values" in raw_result else "resourceData"
    for elt in raw_result[value_to_iter_on]["entries"]:
        key_clean = elt['key'].replace('provider-', '')

        # if the key is not already set before
        if elt["value"] is not None and (key_clean not in data_clean or key_clean == "description"):
            data_clean[key_clean] = parse_key(elt)[1]
    return vra_utils.prettify_key(data_clean)



# -*- coding: utf-8 -*-
import urllib3
import json
import requests
from vra_sdk.vra_config import VraConfig
from vra_sdk.vra_exceptions import VraSdkConfigException, VraSdkRequestException, VraSdkAuthenticateException

class VraAuthenticate():
    """Provide authentication mechanism against vRa and context switching support

    Attributes:
        config (VraConfig): VraConfig object
        login (string): User login
        requestedFor (string): User login + domain
        token (string): vRa token
        domain (string): User ad domain
        environment (string): requested vRa environment
        tenant (string): vRa tenant
    """

    def __init__(self, environment, **kwargs):
        """Init VraAuthenticate
        
        Args:
            environment (string): requested vRa server environment

        Returns:
            VraAuthenticate: self
        """

        urllib3.disable_warnings()
        self.config = VraConfig()
        self.login = None
        self.requestedFor = None
        self.token = None
        self.domain = None
        self.environment = environment

    @property
    def environment(self):
        return self._environment

    @environment.setter
    def environment(self, value):
        """environment property setter

        Will update tenant and vcac_server accordingly
        
        Args:
            value (string): environment as describe in your configuration file
        """

        self.config.vcac_server = self.config.config_file['vcac_servers'][value]
        self.tenant = self.config.config_file['tenant'][value]
        self._environment = value

    def auth_login_password(self, login, password, domain):
        """Managem login/password authentication
        Will update self.token accordingly
        
        Args:
            login (string): vRa login
            password (string): vRa password
            domain (string): AD domain
        
        Raises:
            VraSdkConfigException: Domain requested not declared in configuration file
        
        Returns:
            VraAuthenticate: self
        """

        self.login = login
        self.domain = domain
        self.requestedFor = self.login + "@" + self.domain
        self.token = self.get_token(login, password)
        self.config.session.headers.update(
            {'content-type': 'application/json', 'Accept': 'application/json', 'Authorization': 'Bearer ' + self.token})

        return self

    def auth_login_token(self, login, token, domain):
        """Manage login/token authentication
        Will update self.token accordingly
        
        Args:
            login (string): vRa login
            token (string): vRa token
            domain (string): AD Domain
        
        Raises:
            VraSdkConfigException: Domain requested not declared in configuration file
        
        Returns:
            VraAuthenticate: self
        """

        self.login = login
        self.domain = domain
        self.requestedFor = self.login + "@" + self.domain
        self.token = token
        self.config.session.headers.update(
            {'content-type': 'application/json', 'Accept': 'application/json', 'Authorization': 'Bearer ' + self.token})
        return self

    def get_token(self, login, password):
        """Get authentication token against vRa infrastructure
        
        Args:
            login (string): vRa login
            password (string): vRa password
        
        Raises:
            VraSdkRequestException: Raised if any Requests error
            VraSdkAuthenticateException: Raised for unmanged error. Mostly for raise_for_status 4xx or 5xx errors
        
        Returns:
            string: vRa token
        """

        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json'}
        payload = {
            'username': login,
            'password': password,
            'tenant': self.tenant
        }
        
        try:
            req = self.config.session.post(f"https://{self.config.vcac_server}/identity/api/tokens",
                                           json=payload,
                                           verify=self.config.verify,
                                           headers=headers,
                                           timeout=self.config.timeout)
            req.raise_for_status()
            return json.loads(req.text)['id']
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(f"Error during request to get vRa token: {e}")
        except Exception as e:
            raise VraSdkAuthenticateException(f"Unmanaged error during token retrieving: {e}")

    def delete_token(self):
        """Delete vRa token
        
        Raises:
            VraSdkRequestException: [description]
            VraSdkAuthenticateException: [description]
        
        Returns:
            bolean: True if the token has been succesfully deleted, False there's already no token
        """

        try:
            if not self.token: return False
            req = self.config.session.delete(f"https://{self.config.vcac_server}/identity/api/tokens/{self.token}",
                                           verify=self.config.verify,
                                           headers=self.config.session.headers,
                                           timeout=self.config.timeout)
            req.raise_for_status()
            self.token = None
            self.config.session.headers.update({'content-type':'application/json', 'Accept':'application/json', 'Authorization':''})
            return True
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(f"Error during request to get vRa token: {e}")
        except Exception as e:
            raise VraSdkAuthenticateException(f"Unmanaged error during token retrieving: {e}")



# -*- coding: utf-8 -*-
from vra_sdk.models.vra_object import VraBaseObject
from vra_sdk.vra_config import VraConfig
from vra_sdk.vra_exceptions import VraSdkRequestException, VraSdkPayloadException
from vra_sdk.vra_utils import load_payload_file
import requests
import json
from copy import deepcopy

class BasePayload():
    """Base class for CatalogItem and ResourceAction
    """

    def __init__(self):
        self.config = VraConfig()

    def customize_payload(self, payload, **kwargs):
        """base customization payload
        
        Args:
            payload (dict): base payload, without any customization
        
        Returns:
            dict: customized payload
        """

        if kwargs['payload_type'] == 'CatalogItem':
            payload['catalogItemRef']['id'] = kwargs['catalog_item_id']
            payload['requestedFor'] = kwargs['requested_for']
        else:
            payload['resourceRef']['id'] = kwargs['resource_id']
            payload['resourceActionRef']['id'] = kwargs['resource_action_id']

        payload['organization']['tenantRef'] = kwargs['tenant_name']
        payload['organization']['tenantLabel'] = kwargs['tenant_name']
        payload['organization']['subtenantRef'] = kwargs['business_group_id']
        payload['organization']['subtenantLabel'] = kwargs['business_group_name']

        for k, v in kwargs.items():
            if k not in self.config.config_file['not_in_data']:
                provider_k = k
                if not k.startswith('provider-'):
                    provider_k = 'provider-' + provider_k
                if isinstance(v, bool):
                    payload['requestData']['entries'].append(
                        {"key": provider_k, "value": {"type": "boolean", "value": v}})
                elif isinstance(v, str):
                    payload['requestData']['entries'].append(
                        {"key": provider_k, "value": {"type": "string", "value": v}})
                elif isinstance(v, (int, float)):
                    payload['requestData']['entries'].append(
                        {"key": provider_k, "value": {"type": "decimal", "value": v}})
        return payload


class CatalogItem(BasePayload):

    def __init__(self, customization_func=None, **kwargs):
        """Init the CatalogItem for vRa 6.x payload object
            customization_func ([type], optional): Defaults to None. If not None, this function will add a second customization after the initial one.
        """

        super().__init__()
        self.customized = None
        if not kwargs.get('payload_path'):
            self.base = {
                "@type": "CatalogItemRequest",
                "catalogItemRef": {
                    "id": ""
                },
                "organization": {
                    "tenantRef": "",
                    "subtenantRef": ""
                },
                "requestedFor": "",
                "state": "SUBMITTED",
                "requestNumber": 0,
                "requestData": {
                    "entries": [
                    ]
                }
            }
        else:
            self.base = load_payload_file(kwargs['payload_path'])

        self.customized = self.customize_payload(deepcopy(self.base), **kwargs)
        if customization_func:
            self.customized = customization_func(self.customized, **kwargs)

    def execute_request(self):
        """Execute the request against the vRa infrastructure
        
        Returns:
            requests.Request: request object
        """

        try:
            req = self.config.session.post(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/requests",
                data=json.dumps(self.customized),
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
            return req
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')


class ResourceAction(BasePayload):

    def __init__(self, customization_func=None, **kwargs):
        """Init ResourceAction object for vRa 6.x payload object
            customization_func ([type], optional): Defaults to None. If not None, this function will add a second customization after the initial one.
        """

        super().__init__()
        self.customized = None
        if not kwargs.get('payload_path'):
            self.base = {
                "@type": "ResourceActionRequest",
                "resourceRef": {
                    "id": ""
                },
                "resourceActionRef": {
                    "id": ""
                },
                "organization": {
                    "tenantRef": "",
                    "tenantLabel": "",
                    "subtenantRef": "",
                    "subtenantLabel": ""
                },
                "state": "SUBMITTED",
                "requestNumber": 0,
                "requestData": {
                    "entries": [
                    ]
                }
            }
        else:
            self.base = load_payload_file(kwargs['payload_path'])

        self.customized = self.customize_payload(deepcopy(self.base), **kwargs)
        if customization_func:
            self.customized = customization_func(self.customized, **kwargs)

    def execute_request(self):
        """Execute the request against the vRa infrastructure
        
        Returns:
            requests.Request: request object
        """
        try:
            req = self.config.session.post(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/requests",
                data=json.dumps(self.customized),
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'vRa request exception : {e}')
        except Exception as e:
            raise VraSdkPayloadException(
                f'Unmanaged error requesting vRa: {e}')



# -*- coding: utf-8 -*-
import json
import requests
import re
import os
import os.path
import time
import urllib3
from vra_sdk.vra_request import VraRequest
from vra_sdk import vra_decorator, vra_utils
from vra_sdk.vra_config import VraConfig
from vra_sdk.vra_exceptions import VraSdkMainException, VraSdkRequestException
from vra_sdk.vra_factory import VraFactory


class VraSdk():
    """Core class of the library.

    Attributes:
        config (VraConfig): VraConfig object
        business_group (string): current business group name
        business_group_id (string): current business group id
        catalog (dict): map of catalog item/resource action to related vRa id
    """

    def __init__(self, authentication_object, business_group, **kwargs):
        """Init the VraSdk object
        
        Args:
            authentication_object (VraAuthenticate): authentication object
            business_group (string): business group to work on
        """

        urllib3.disable_warnings()
        self.authentication_object = authentication_object
        self.config = VraConfig()
        self.business_group_id = ''
        self.business_group = business_group
        self.catalog = self.get_catalog()

    @property
    def business_group(self):
        """property to get business group name"""
        return self._business_group

    @business_group.setter
    def business_group(self, value):
        """property to set business group name and update the business group id
        
        Args:
            value (string): business group name to set
        """

        self.business_group_id = self.get_bg_id(value)
        self._business_group = value

    def get_bg_id(self, business_group, force_refresh=False):
        """get business group id against vRa infrastructure
        
        Args:
            business_group (string): business group name
            force_refresh (bool, optional): Defaults to False. force the refresh of the business group id
        
        """

        if not self.business_group_id or force_refresh:
            try:
                req = self.config.session.get(
                    f"https://{self.config.vcac_server}/catalog-service/api/consumer/entitledCatalogItems?limit=998",
                    verify=self.config.verify, timeout=self.config.timeout)
                req.raise_for_status()
                response = json.loads(req.text)
            except requests.exceptions.RequestException as e:
                raise VraSdkRequestException(
                    f"Error getting business group id for bg {self.business_group}: {e}")
            except Exception as e:
                raise VraSdkMainException(e)

            if 'content' in response:
                for catalog_item in response['content']:
                    for elt in catalog_item['entitledOrganizations']:
                        if elt["subtenantLabel"] == business_group:
                            return elt['subtenantRef']
                raise VraSdkMainException(
                    f'No entitlement for the account {self.authentication_object.login} in business group {business_group}')
            else:
                raise VraSdkMainException(
                    f'Unable get bg id list. No entitled catalog item for account {self.authentication_object.login}')

    def get_catalog(self):
        """Get the catalog item list from the vRa infrastructure
        
        Returns:
            dict: map of catalog item to vRa id
        """

        catalog = {}
        try:
            req = self.config.session.get(
                f"https://{self.config.vcac_server}/catalog-service/api/consumer/entitledCatalogItems?limit=9999",
                verify=self.config.verify,
                timeout=self.config.timeout)
            req.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise VraSdkRequestException(
                f'Error during retrieving catalog item: {e}')
        except Exception as e:
            raise VraSdkMainException(
                f'Unmanaged error during catalog retrieving: {e}')

        try:
            for elt in json.loads(req.text)['content']:
                catalog[elt['catalogItem']['name']] = elt['catalogItem']['id']
        except Exception as e:
            raise VraSdkMainException(
                f'Error updating the catalog attribute {e}')

        return catalog

    @vra_decorator.check_entitlement
    def request_catalog_item(self, item_name, customization_func=None, **kwargs):
        """create a ready to use Request object to request a catalog item
        
        Args:
            item_name (string): catalog item name
            customization_func (function, optional): Defaults to None. payload customization function
        
        Returns:
            VraRequest: object with the payload attribute well customized
        """

        if not kwargs.get('payload_version'):
            kwargs['payload_version'] = self.config.config_file['payload_default_version']
        payload = self.format_payload(
            'catalog_item', item_name, customization_func, **kwargs)
        return VraRequest(payload)

    @vra_decorator.update_catalog_resource_operation
    @vra_decorator.check_entitlement
    def request_resource_action(self, action_name, resource_id, customization_func=None, **kwargs):
        """create a ready to use Request object to request a resource action
        
        Args:
            action_name (string): resource action name
            resource_id (string): resource id to perform action on
            customization_func (function, optional): Defaults to None. payload customization function
        
        Returns:
            VraRequest: object with the payload attribute well customized
        """
        if not kwargs.get('payload_version'):
            kwargs['payload_version'] = self.config.config_file['payload_default_version']
        payload = self.format_payload(
            'resource_action', action_name, resource_id, customization_func, **kwargs)
        return VraRequest(payload)

    def format_payload(self, origin, *args, **kwargs):
        """Customized request kwargs before using it in the VraFactory class to create a payload object
        
        Args:
            origin (string): origin of the request. 'catalog_item' or 'resource_action'
        
        Returns:
            object: CatalogItem or ResourceAction object from one of the vra_payload_x modulme
        """

        kwargs['requested_for'] = self.authentication_object.requestedFor
        kwargs['business_group_name'] = self.business_group
        kwargs['business_group_id'] = self.business_group_id
        kwargs['tenant_name'] = self.authentication_object.tenant

        if args[0] in self.config.config_file[origin] and self.config.config_file[origin][args[0]].get('payload'):
            kwargs['payload_path'] = self.config.config_file[origin][args[0]].get(
                'payload')

        if origin == 'catalog_item':
            kwargs['catalog_item_name'] = args[0]
            kwargs['catalog_item_id'] = self.catalog[args[0]]
            kwargs['payload_type'] = 'CatalogItem'
            customization_func = args[1]
        elif origin == 'resource_action':
            kwargs['resource_id'] = args[1]
            kwargs['resource_action_name'] = args[0]
            kwargs['resource_action_id'] = self.catalog[args[0]]
            kwargs['payload_type'] = 'ResourceAction'
            customization_func = args[2]

        return VraFactory.factory('payload', customization_func, **kwargs)

    def get_data(self, object_type, key, value):
        """Get data about one catalog resource in vRa. Get detailed info about your object
        
        Args:
            object_type (string): object type as described in the 'business_models' section of the configuration fiel
            key (string): field to filter on
            value (string): value of the field
        
        Returns:
            object: business models object type as described in you configuration file
        """

        data = VraRequest({}).get_object(object_type, key, value, 1,1, True)

        if not data:
            raise VraSdkMainException(
                f'No {object_type} exist with {key}={value}')

        return data[0]

    def list_data(self, object_type, key, value, limit=None, page=1, full=False, recursive=False):
        """Get info about a list of object. Get less details than get_data(), but you still get the id
        
        Args:
            object_type (string): object typre
            key (string): field to filter on
            value (string): value of the field to search on
            limit (int, optional): Defaults to None. maximum result
            page (int, optional): Defaults to 1. vRa result page to get data from
            recursive (bool, optional): Defaults to False. If True, will call itself recursively until there's no more page to get from
        
        Raises:
            VraSdkMainException: [description]
        
        Returns:
            list: list of business models object type as described in you configuration file
        """
        if not limit:
            limit = self.config.config_file['max_vra_result_per_page']

        data = VraRequest({}).get_object(object_type, key, value, limit, page, full)
        if not data:
            raise VraSdkMainException(f'No {object_type} exist with {key}={value}')

        if recursive and (len(data) == limit):
            page = page+1
            data_to_add = self.list_data(object_type, key, value, limit, page, full, recursive)
            data = data + data_to_add
        return data

    def get_raw_definition(self, key, value, resource_type):
        """Return dict used to create object inside the VraFactory

        This methods is a helper. You should use only at the beginning to help you build the object definitions needed for the factrory

        Args:
            key (string): key to search data from
            value (string): value of the key
            resource_type (string): vRa resource type

        Returns:
            dict: dict of user friendly vRa formatted data
        """
        
        return VraRequest({}).get_object(None, key, value, 1, 1, True, resource_type)[0]

